"""Utility functions"""

from typing import Any, Dict, List
import re


def snake_case(name: str) -> str:
    """Convert CamelCase to snake_case"""
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def dict_to_model(data: Dict[str, Any], model_class) -> Any:
    """Convert dictionary to model instance"""
    return model_class(**data)


def models_to_dict(models: List[Any]) -> List[Dict[str, Any]]:
    """Convert list of models to list of dictionaries"""
    return [model.to_dict() for model in models]


"""Database migration system"""

import os
import json
from datetime import datetime
from typing import List, Dict, Any, Optional
from pathlib import Path

from .database import Database
from .models import Model
from .exceptions import MigrationError


class Migration:
    """Single migration"""

    def __init__(self, name: str, operations: List[Dict[str, Any]]):
        self.name = name
        self.operations = operations
        self.applied_at = None

    async def apply(self, database: Database):
        """Apply migration"""
        for operation in self.operations:
            await self._execute_operation(database, operation)

    async def _execute_operation(self, database: Database, operation: Dict[str, Any]):
        """Execute single operation"""
        op_type = operation.get('type')
        
        if op_type == 'create_table':
            await self._create_table(database, operation)
        elif op_type == 'drop_table':
            await self._drop_table(database, operation)
        elif op_type == 'add_column':
            await self._add_column(database, operation)
        elif op_type == 'drop_column':
            await self._drop_column(database, operation)
        elif op_type == 'raw_sql':
            await database.execute(operation['sql'])

    async def _create_table(self, database: Database, operation: Dict[str, Any]):
        """Create table operation"""
        model = operation['model']
        await model.create_table()

    async def _drop_table(self, database: Database, operation: Dict[str, Any]):
        """Drop table operation"""
        table_name = operation['table_name']
        await database.execute(f"DROP TABLE IF EXISTS {table_name}")

    async def _add_column(self, database: Database, operation: Dict[str, Any]):
        """Add column operation"""
        table_name = operation['table_name']
        column_name = operation['column_name']
        column_type = operation['column_type']
        
        query = f"ALTER TABLE {table_name} ADD COLUMN {column_name} {column_type}"
        await database.execute(query)

    async def _drop_column(self, database: Database, operation: Dict[str, Any]):
        """Drop column operation"""
        table_name = operation['table_name']
        column_name = operation['column_name']
        
        # SQLite doesn't support DROP COLUMN directly
        database_type = database.__class__.__name__
        if 'Sqlite' in database_type:
            raise MigrationError("SQLite doesn't support DROP COLUMN. Use raw SQL with table recreation.")
        
        query = f"ALTER TABLE {table_name} DROP COLUMN {column_name}"
        await database.execute(query)


class MigrationManager:
    """Migration manager"""

    def __init__(self, database: Database, migrations_dir: str = "migrations"):
        self.database = database
        self.migrations_dir = Path(migrations_dir)
        self.migrations_dir.mkdir(exist_ok=True)
        self._migrations: List[Migration] = []

    async def init(self):
        """Initialize migration system"""
        await self._create_migrations_table()

    async def _create_migrations_table(self):
        """Create migrations tracking table"""
        query = """
        CREATE TABLE IF NOT EXISTS migrations (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name VARCHAR(255) UNIQUE NOT NULL,
            applied_at DATETIME NOT NULL
        )
        """
        await self.database.execute(query)

    def create_migration(self, name: str, operations: List[Dict[str, Any]]) -> Migration:
        """Create a new migration"""
        timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        migration_name = f"{timestamp}_{name}"
        
        migration = Migration(migration_name, operations)
        
        # Save to file
        migration_file = self.migrations_dir / f"{migration_name}.json"
        with open(migration_file, 'w') as f:
            json.dump({
                'name': migration_name,
                'operations': operations
            }, f, indent=2)
        
        return migration

    def load_migrations(self) -> List[Migration]:
        """Load all migrations from directory"""
        migrations = []
        
        for file_path in sorted(self.migrations_dir.glob("*.json")):
            with open(file_path, 'r') as f:
                data = json.load(f)
                migration = Migration(data['name'], data['operations'])
                migrations.append(migration)
        
        return migrations

    async def get_applied_migrations(self) -> List[str]:
        """Get list of applied migrations"""
        rows = await self.database.fetch_all("SELECT name FROM migrations ORDER BY applied_at")
        return [row['name'] for row in rows]

    async def migrate(self):
        """Apply pending migrations"""
        all_migrations = self.load_migrations()
        applied = await self.get_applied_migrations()
        
        pending = [m for m in all_migrations if m.name not in applied]
        
        for migration in pending:
            print(f"Applying migration: {migration.name}")
            await migration.apply(self.database)
            
            # Record migration
            placeholder = self.database.get_placeholder()
            await self.database.execute(
                f"INSERT INTO migrations (name, applied_at) VALUES ({placeholder}, {placeholder})",
                (migration.name, datetime.now().isoformat())
            )
            print(f"✓ Applied: {migration.name}")

    async def rollback(self, steps: int = 1):
        """Rollback migrations"""
        applied = await self.get_applied_migrations()
        
        if not applied:
            print("No migrations to rollback")
            return
        
        to_rollback = applied[-steps:]
        
        for migration_name in reversed(to_rollback):
            print(f"Rolling back: {migration_name}")
            placeholder = self.database.get_placeholder()
            await self.database.execute(
                f"DELETE FROM migrations WHERE name = {placeholder}",
                (migration_name,)
            )
            print(f"✓ Rolled back: {migration_name}")

    def make_migrations(self, models: List[type]):
        """Auto-generate migrations from models"""
        operations = []
        
        for model in models:
            if not issubclass(model, Model):
                continue
            
            operations.append({
                'type': 'create_table',
                'model': model
            })
        
        if operations:
            migration = self.create_migration("auto_generated", operations)
            print(f"Created migration: {migration.name}")
            return migration
        else:
            print("No changes detected")
            return None


"""Database connection and management"""

import asyncio
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple
from contextlib import asynccontextmanager

try:
    import aiosqlite
except ImportError:
    aiosqlite = None

try:
    import asyncmy
except ImportError:
    asyncmy = None

try:
    import psycopg
    from psycopg.rows import dict_row
except ImportError:
    psycopg = None

from .exceptions import DatabaseError, ConfigurationError


class Database(ABC):
    """Abstract base class for database connections"""

    def __init__(self, database: str, **kwargs):
        self.database = database
        self.connection_params = kwargs
        self._connection = None
        self._is_connected = False
        self._pool = None

    @abstractmethod
    async def connect(self):
        """Establish database connection"""
        pass

    @abstractmethod
    async def close(self):
        """Close database connection"""
        pass

    @abstractmethod
    async def execute(self, query: str, params: Optional[tuple] = None) -> Any:
        """Execute a query"""
        pass

    @abstractmethod
    async def fetch_one(self, query: str, params: Optional[tuple] = None) -> Optional[Dict]:
        """Fetch one row"""
        pass

    @abstractmethod
    async def fetch_all(self, query: str, params: Optional[tuple] = None) -> List[Dict]:
        """Fetch all rows"""
        pass

    @abstractmethod
    async def execute_many(self, query: str, params_list: List[tuple]) -> None:
        """Execute query multiple times"""
        pass

    @abstractmethod
    def get_placeholder(self) -> str:
        """Get parameter placeholder for this database"""
        pass

    @abstractmethod
    async def get_last_insert_id(self) -> int:
        """Get last inserted ID"""
        pass

    @abstractmethod
    async def begin_transaction(self):
        """Begin a transaction"""
        pass

    @abstractmethod
    async def commit(self):
        """Commit transaction"""
        pass

    @abstractmethod
    async def rollback(self):
        """Rollback transaction"""
        pass

    @asynccontextmanager
    async def transaction(self):
        """Context manager for transactions"""
        await self.begin_transaction()
        try:
            yield
            await self.commit()
        except Exception:
            await self.rollback()
            raise

    async def create_tables(self, models: List['Model']):
        """Create tables for given models"""
        for model in models:
            await model.create_table()

    async def drop_tables(self, models: List['Model']):
        """Drop tables for given models"""
        for model in models:
            await model.drop_table()

    @property
    def is_connected(self) -> bool:
        """Check if database is connected"""
        return self._is_connected


class SqliteDatabase(Database):
    """SQLite database implementation"""

    def __init__(self, database: str, **kwargs):
        if aiosqlite is None:
            raise ConfigurationError("aiosqlite is not installed. Install it with: pip install aiosqlite")
        super().__init__(database, **kwargs)

    async def connect(self):
        """Connect to SQLite database"""
        if not self._is_connected:
            self._connection = await aiosqlite.connect(self.database, **self.connection_params)
            self._connection.row_factory = aiosqlite.Row
            self._is_connected = True

    async def close(self):
        """Close SQLite connection"""
        if self._connection:
            await self._connection.close()
            self._is_connected = False

    async def execute(self, query: str, params: Optional[tuple] = None) -> Any:
        """Execute query"""
        if not self._is_connected:
            await self.connect()
        cursor = await self._connection.execute(query, params or ())
        await self._connection.commit()
        return cursor

    async def fetch_one(self, query: str, params: Optional[tuple] = None) -> Optional[Dict]:
        """Fetch one row"""
        if not self._is_connected:
            await self.connect()
        cursor = await self._connection.execute(query, params or ())
        row = await cursor.fetchone()
        return dict(row) if row else None

    async def fetch_all(self, query: str, params: Optional[tuple] = None) -> List[Dict]:
        """Fetch all rows"""
        if not self._is_connected:
            await self.connect()
        cursor = await self._connection.execute(query, params or ())
        rows = await cursor.fetchall()
        return [dict(row) for row in rows]

    async def execute_many(self, query: str, params_list: List[tuple]) -> None:
        """Execute many"""
        if not self._is_connected:
            await self.connect()
        await self._connection.executemany(query, params_list)
        await self._connection.commit()

    def get_placeholder(self) -> str:
        """Get placeholder"""
        return "?"

    async def get_last_insert_id(self) -> int:
        """Get last insert ID"""
        cursor = await self._connection.execute("SELECT last_insert_rowid()")
        row = await cursor.fetchone()
        return row[0]

    async def begin_transaction(self):
        """Begin transaction"""
        await self._connection.execute("BEGIN")

    async def commit(self):
        """Commit transaction"""
        await self._connection.commit()

    async def rollback(self):
        """Rollback transaction"""
        await self._connection.rollback()


class MySQLDatabase(Database):
    """MySQL database implementation"""

    def __init__(self, database: str, host: str = "localhost", port: int = 3306,
                 user: str = "root", password: str = "", **kwargs):
        if asyncmy is None:
            raise ConfigurationError("asyncmy is not installed. Install it with: pip install asyncmy")
        super().__init__(database, **kwargs)
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    async def connect(self):
        """Connect to MySQL"""
        if not self._is_connected:
            self._connection = await asyncmy.connect(
                host=self.host,
                port=self.port,
                user=self.user,
                password=self.password,
                db=self.database,
                **self.connection_params
            )
            self._is_connected = True

    async def close(self):
        """Close connection"""
        if self._connection:
            self._connection.close()
            await self._connection.wait_closed()
            self._is_connected = False

    async def execute(self, query: str, params: Optional[tuple] = None) -> Any:
        """Execute query"""
        if not self._is_connected:
            await self.connect()
        async with self._connection.cursor() as cursor:
            await cursor.execute(query, params or ())
            await self._connection.commit()
            return cursor

    async def fetch_one(self, query: str, params: Optional[tuple] = None) -> Optional[Dict]:
        """Fetch one row"""
        if not self._is_connected:
            await self.connect()
        async with self._connection.cursor(asyncmy.DictCursor) as cursor:
            await cursor.execute(query, params or ())
            return await cursor.fetchone()

    async def fetch_all(self, query: str, params: Optional[tuple] = None) -> List[Dict]:
        """Fetch all rows"""
        if not self._is_connected:
            await self.connect()
        async with self._connection.cursor(asyncmy.DictCursor) as cursor:
            await cursor.execute(query, params or ())
            return await cursor.fetchall()

    async def execute_many(self, query: str, params_list: List[tuple]) -> None:
        """Execute many"""
        if not self._is_connected:
            await self.connect()
        async with self._connection.cursor() as cursor:
            await cursor.executemany(query, params_list)
            await self._connection.commit()

    def get_placeholder(self) -> str:
        """Get placeholder"""
        return "%s"

    async def get_last_insert_id(self) -> int:
        """Get last insert ID"""
        async with self._connection.cursor() as cursor:
            await cursor.execute("SELECT LAST_INSERT_ID()")
            row = await cursor.fetchone()
            return row[0]

    async def begin_transaction(self):
        """Begin transaction"""
        await self._connection.begin()

    async def commit(self):
        """Commit"""
        await self._connection.commit()

    async def rollback(self):
        """Rollback"""
        await self._connection.rollback()


class PostgreSQLDatabase(Database):
    """PostgreSQL database implementation"""

    def __init__(self, database: str, host: str = "localhost", port: int = 5432,
                 user: str = "postgres", password: str = "", **kwargs):
        if psycopg is None:
            raise ConfigurationError("psycopg is not installed. Install it with: pip install psycopg[binary]")
        super().__init__(database, **kwargs)
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    async def connect(self):
        """Connect to PostgreSQL"""
        if not self._is_connected:
            conninfo = f"host={self.host} port={self.port} dbname={self.database} user={self.user} password={self.password}"
            self._connection = await psycopg.AsyncConnection.connect(conninfo, row_factory=dict_row)
            self._is_connected = True

    async def close(self):
        """Close connection"""
        if self._connection:
            await self._connection.close()
            self._is_connected = False

    async def execute(self, query: str, params: Optional[tuple] = None) -> Any:
        """Execute query"""
        if not self._is_connected:
            await self.connect()
        async with self._connection.cursor() as cursor:
            await cursor.execute(query, params or ())
            await self._connection.commit()
            return cursor

    async def fetch_one(self, query: str, params: Optional[tuple] = None) -> Optional[Dict]:
        """Fetch one row"""
        if not self._is_connected:
            await self.connect()
        async with self._connection.cursor() as cursor:
            await cursor.execute(query, params or ())
            return await cursor.fetchone()

    async def fetch_all(self, query: str, params: Optional[tuple] = None) -> List[Dict]:
        """Fetch all rows"""
        if not self._is_connected:
            await self.connect()
        async with self._connection.cursor() as cursor:
            await cursor.execute(query, params or ())
            return await cursor.fetchall()

    async def execute_many(self, query: str, params_list: List[tuple]) -> None:
        """Execute many"""
        if not self._is_connected:
            await self.connect()
        async with self._connection.cursor() as cursor:
            for params in params_list:
                await cursor.execute(query, params)
            await self._connection.commit()

    def get_placeholder(self) -> str:
        """Get placeholder"""
        return "%s"

    async def get_last_insert_id(self) -> int:
        """Get last insert ID (PostgreSQL uses RETURNING clause)"""
        # This is handled in the insert query itself
        return 0

    async def begin_transaction(self):
        """Begin transaction"""
        await self._connection.execute("BEGIN")

    async def commit(self):
        """Commit"""
        await self._connection.commit()

    async def rollback(self):
        """Rollback"""
        await self._connection.rollback()


"""Custom exceptions for UniORM"""


class UniORMException(Exception):
    """Base exception for UniORM"""
    pass


class DatabaseError(UniORMException):
    """Database operation error"""
    pass


class IntegrityError(DatabaseError):
    """Data integrity error (e.g., unique constraint violation)"""
    pass


class DoesNotExist(UniORMException):
    """Object does not exist in database"""
    pass


class MultipleObjectsReturned(UniORMException):
    """Multiple objects returned when one was expected"""
    pass


class ConfigurationError(UniORMException):
    """Configuration error"""
    pass


class MigrationError(UniORMException):
    """Migration error"""
    pass


"""
UniORM - Universal Async ORM
A powerful async ORM supporting SQLite, MySQL, and PostgreSQL
"""

__version__ = "1.0.0"
__author__ = "oscoderuz"

from .database import (
    Database,
    SqliteDatabase,
    MySQLDatabase,
    PostgreSQLDatabase,
)
from .fields import (
    Field,
    CharField,
    TextField,
    IntField,
    BigIntField,
    FloatField,
    BoolField,
    DateField,
    DateTimeField,
    TimeField,
    JSONField,
    ForeignKeyField,
)
from .models import Model
from .query import Q
from .exceptions import (
    UniORMException,
    DatabaseError,
    IntegrityError,
    DoesNotExist,
    MultipleObjectsReturned,
)
from .migrations import MigrationManager

__all__ = [
    # Database
    "Database",
    "SqliteDatabase",
    "MySQLDatabase",
    "PostgreSQLDatabase",
    # Fields
    "Field",
    "CharField",
    "TextField",
    "IntField",
    "BigIntField",
    "FloatField",
    "BoolField",
    "DateField",
    "DateTimeField",
    "TimeField",
    "JSONField",
    "ForeignKeyField",
    # Models
    "Model",
    # Query
    "Q",
    # Exceptions
    "UniORMException",
    "DatabaseError",
    "IntegrityError",
    "DoesNotExist",
    "MultipleObjectsReturned",
    # Migrations
    "MigrationManager",
]


"""Query builder for UniORM"""

from typing import Any, Dict, List, Optional, Union, TYPE_CHECKING
from enum import Enum

if TYPE_CHECKING:
    from .models import Model


class JoinType(Enum):
    """Join types"""
    INNER = "INNER JOIN"
    LEFT = "LEFT JOIN"
    RIGHT = "RIGHT JOIN"
    FULL = "FULL OUTER JOIN"


class Q:
    """Query condition builder"""

    def __init__(self, **kwargs):
        self.conditions = kwargs
        self.connector = "AND"
        self.negated = False
        self.children = []

    def __and__(self, other):
        """AND operation"""
        q = Q()
        q.connector = "AND"
        q.children = [self, other]
        return q

    def __or__(self, other):
        """OR operation"""
        q = Q()
        q.connector = "OR"
        q.children = [self, other]
        return q

    def __invert__(self):
        """NOT operation"""
        q = Q(**self.conditions)
        q.negated = not self.negated
        q.children = self.children
        q.connector = self.connector
        return q


class QueryBuilder:
    """SQL query builder"""

    def __init__(self, model: 'Model'):
        self.model = model
        self._where_conditions = []
        self._where_params = []
        self._order_by = []
        self._limit_value = None
        self._offset_value = None
        self._select_fields = []
        self._joins = []
        self._group_by = []
        self._having_conditions = []

    def where(self, *args, **kwargs) -> 'QueryBuilder':
        """Add WHERE conditions"""
        if args:
            for arg in args:
                if isinstance(arg, Q):
                    self._process_q_object(arg)
        
        if kwargs:
            for key, value in kwargs.items():
                self._add_condition(key, value)
        
        return self

    def _add_condition(self, key: str, value: Any):
        """Add a single condition"""
        parts = key.split('__')
        field_name = parts[0]
        operator = parts[1] if len(parts) > 1 else 'exact'

        placeholder = self.model._meta.database.get_placeholder()

        if operator == 'exact':
            self._where_conditions.append(f"{field_name} = {placeholder}")
            self._where_params.append(value)
        elif operator == 'gt':
            self._where_conditions.append(f"{field_name} > {placeholder}")
            self._where_params.append(value)
        elif operator == 'gte':
            self._where_conditions.append(f"{field_name} >= {placeholder}")
            self._where_params.append(value)
        elif operator == 'lt':
            self._where_conditions.append(f"{field_name} < {placeholder}")
            self._where_params.append(value)
        elif operator == 'lte':
            self._where_conditions.append(f"{field_name} <= {placeholder}")
            self._where_params.append(value)
        elif operator == 'ne':
            self._where_conditions.append(f"{field_name} != {placeholder}")
            self._where_params.append(value)
        elif operator == 'in':
            placeholders = ', '.join([placeholder] * len(value))
            self._where_conditions.append(f"{field_name} IN ({placeholders})")
            self._where_params.extend(value)
        elif operator == 'contains':
            self._where_conditions.append(f"{field_name} LIKE {placeholder}")
            self._where_params.append(f"%{value}%")
        elif operator == 'startswith':
            self._where_conditions.append(f"{field_name} LIKE {placeholder}")
            self._where_params.append(f"{value}%")
        elif operator == 'endswith':
            self._where_conditions.append(f"{field_name} LIKE {placeholder}")
            self._where_params.append(f"%{value}")
        elif operator == 'isnull':
            if value:
                self._where_conditions.append(f"{field_name} IS NULL")
            else:
                self._where_conditions.append(f"{field_name} IS NOT NULL")

    def _process_q_object(self, q: Q):
        """Process Q object"""
        if q.conditions:
            for key, value in q.conditions.items():
                self._add_condition(key, value)

    def order_by(self, *fields: str) -> 'QueryBuilder':
        """Add ORDER BY clause"""
        for field in fields:
            if field.startswith('-'):
                self._order_by.append(f"{field[1:]} DESC")
            else:
                self._order_by.append(f"{field} ASC")
        return self

    def limit(self, value: int) -> 'QueryBuilder':
        """Add LIMIT clause"""
        self._limit_value = value
        return self

    def offset(self, value: int) -> 'QueryBuilder':
        """Add OFFSET clause"""
        self._offset_value = value
        return self

    def join(self, model: 'Model', on: str, join_type: JoinType = JoinType.INNER) -> 'QueryBuilder':
        """Add JOIN clause"""
        self._joins.append({
            'model': model,
            'on': on,
            'type': join_type
        })
        return self

    def group_by(self, *fields: str) -> 'QueryBuilder':
        """Add GROUP BY clause"""
        self._group_by.extend(fields)
        return self

    def build_select(self) -> tuple:
        """Build SELECT query"""
        table_name = self.model._meta.table_name
        
        # SELECT clause
        if self._select_fields:
            select_clause = f"SELECT {', '.join(self._select_fields)}"
        else:
            select_clause = f"SELECT * FROM {table_name}"

        # JOIN clauses
        join_clause = ""
        if self._joins:
            join_parts = []
            for join in self._joins:
                join_type = join['type'].value
                join_table = join['model']._meta.table_name
                join_parts.append(f"{join_type} {join_table} ON {join['on']}")
            join_clause = " " + " ".join(join_parts)

        # WHERE clause
        where_clause = ""
        if self._where_conditions:
            where_clause = " WHERE " + " AND ".join(self._where_conditions)

        # GROUP BY clause
        group_by_clause = ""
        if self._group_by:
            group_by_clause = " GROUP BY " + ", ".join(self._group_by)

        # ORDER BY clause
        order_by_clause = ""
        if self._order_by:
            order_by_clause = " ORDER BY " + ", ".join(self._order_by)

        # LIMIT/OFFSET clause
        limit_clause = ""
        if self._limit_value is not None:
            limit_clause = f" LIMIT {self._limit_value}"
        if self._offset_value is not None:
            limit_clause += f" OFFSET {self._offset_value}"

        query = f"{select_clause}{join_clause}{where_clause}{group_by_clause}{order_by_clause}{limit_clause}"
        return query, tuple(self._where_params)

    async def execute(self) -> List['Model']:
        """Execute query and return results"""
        query, params = self.build_select()
        rows = await self.model._meta.database.fetch_all(query, params)
        return [self.model._from_db(row) for row in rows]

    async def first(self) -> Optional['Model']:
        """Get first result"""
        self.limit(1)
        results = await self.execute()
        return results[0] if results else None

    async def count(self) -> int:
        """Count results"""
        table_name = self.model._meta.table_name
        where_clause = ""
        if self._where_conditions:
            where_clause = " WHERE " + " AND ".join(self._where_conditions)
        
        query = f"SELECT COUNT(*) as count FROM {table_name}{where_clause}"
        result = await self.model._meta.database.fetch_one(query, tuple(self._where_params))
        return result['count'] if result else 0

    async def exists(self) -> bool:
        """Check if any results exist"""
        return await self.count() > 0

    async def delete(self) -> int:
        """Delete matching records"""
        table_name = self.model._meta.table_name
        where_clause = ""
        if self._where_conditions:
            where_clause = " WHERE " + " AND ".join(self._where_conditions)
        
        query = f"DELETE FROM {table_name}{where_clause}"
        await self.model._meta.database.execute(query, tuple(self._where_params))
        return await self.count()

    async def update(self, **kwargs) -> int:
        """Update matching records"""
        if not kwargs:
            return 0

        table_name = self.model._meta.table_name
        placeholder = self.model._meta.database.get_placeholder()
        
        set_parts = []
        params = []
        for key, value in kwargs.items():
            set_parts.append(f"{key} = {placeholder}")
            params.append(value)
        
        params.extend(self._where_params)
        
        set_clause = ", ".join(set_parts)
        where_clause = ""
        if self._where_conditions:
            where_clause = " WHERE " + " AND ".join(self._where_conditions)
        
        query = f"UPDATE {table_name} SET {set_clause}{where_clause}"
        await self.model._meta.database.execute(query, tuple(params))
        return await self.count()


from setuptools import setup, find_packages

setup(
    name="uniorm",
    version="1.0.0",
    author="oscoderuz",
    author_email="oscoderuz@gmail.com",
    description="Universal Async ORM for SQLite, MySQL, and PostgreSQL",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/oscoderuz/uniorm",
    packages=find_packages(),
    python_requires=">=3.9",
    install_requires=[],
    extras_require={
        "sqlite": ["aiosqlite"],
        "mysql": ["asyncmy"],
        "postgres": ["psycopg[binary]"],
        "all": [
            "aiosqlite",
            "asyncmy",
            "psycopg[binary]"
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Framework :: AsyncIO",
        "Topic :: Database :: Front-Ends",
    ],
)


"""Model base class and metaclass"""

from typing import Any, Dict, List, Optional, Type, ClassVar
from datetime import datetime

from .fields import Field, ForeignKeyField, DateTimeField, DateField
from .query import QueryBuilder, Q
from .exceptions import DoesNotExist, MultipleObjectsReturned, ConfigurationError


class ModelOptions:
    """Model metadata"""

    def __init__(self, meta=None):
        self.table_name = None
        self.database = None
        self.indexes = []
        self.ordering = []
        
        if meta:
            for attr in dir(meta):
                if not attr.startswith('_'):
                    setattr(self, attr, getattr(meta, attr))


class ModelMeta(type):
    """Metaclass for Model"""

    def __new__(mcs, name, bases, attrs):
        # Don't process Model class itself
        if name == 'Model':
            return super().__new__(mcs, name, bases, attrs)

        # Collect fields
        fields = {}
        for key, value in list(attrs.items()):
            if isinstance(value, Field):
                value.name = key
                fields[key] = value

        # Get Meta options
        meta = attrs.pop('Meta', None)
        options = ModelOptions(meta)

        # Set table name
        if not options.table_name:
            options.table_name = name.lower() + 's'

        # Create new class
        new_class = super().__new__(mcs, name, bases, attrs)
        new_class._meta = options
        new_class._fields = fields

        # Set model reference in fields
        for field in fields.values():
            field.model = new_class

        return new_class


class Model(metaclass=ModelMeta):
    """Base model class"""

    _meta: ClassVar[ModelOptions]
    _fields: ClassVar[Dict[str, Field]]

    def __init__(self, **kwargs):
        self._data = {}
        self._is_saved = False

        # Set primary key
        if 'id' not in self._fields:
            from .fields import IntField
            id_field = IntField(primary_key=True, auto_increment=True)
            id_field.name = 'id'
            id_field.model = self.__class__
            self._fields['id'] = id_field

        # Initialize fields with defaults
        for name, field in self._fields.items():
            if name in kwargs:
                setattr(self, name, kwargs[name])
            elif field.default is not None:
                setattr(self, name, field.get_default())
            else:
                self._data[name] = None

    def __setattr__(self, name: str, value: Any):
        if name.startswith('_'):
            super().__setattr__(name, value)
        elif name in self._fields:
            self._data[name] = value
        else:
            super().__setattr__(name, value)

    def __getattribute__(self, name: str):
        if name.startswith('_') or name in ('save', 'delete', 'refresh', 'to_dict'):
            return super().__getattribute__(name)
        
        fields = super().__getattribute__('_fields')
        if name in fields:
            data = super().__getattribute__('_data')
            return data.get(name)
        
        return super().__getattribute__(name)

    @classmethod
    def select(cls, *fields) -> QueryBuilder:
        """Create a SELECT query"""
        query = QueryBuilder(cls)
        if fields:
            query._select_fields = list(fields)
        return query

    @classmethod
    async def get(cls, **kwargs) -> 'Model':
        """Get a single object"""
        results = await cls.select().where(**kwargs).execute()
        
        if not results:
            raise DoesNotExist(f"{cls.__name__} matching query does not exist")
        
        if len(results) > 1:
            raise MultipleObjectsReturned(f"get() returned more than one {cls.__name__}")
        
        return results[0]

    @classmethod
    async def get_or_none(cls, **kwargs) -> Optional['Model']:
        """Get a single object or None"""
        try:
            return await cls.get(**kwargs)
        except DoesNotExist:
            return None

    @classmethod
    async def create(cls, **kwargs) -> 'Model':
        """Create and save a new object"""
        instance = cls(**kwargs)
        await instance.save()
        return instance

    @classmethod
    async def get_or_create(cls, defaults=None, **kwargs) -> tuple['Model', bool]:
        """Get or create an object"""
        try:
            instance = await cls.get(**kwargs)
            return instance, False
        except DoesNotExist:
            create_kwargs = kwargs.copy()
            if defaults:
                create_kwargs.update(defaults)
            instance = await cls.create(**create_kwargs)
            return instance, True

    @classmethod
    async def filter(cls, **kwargs) -> List['Model']:
        """Filter objects"""
        return await cls.select().where(**kwargs).execute()

    @classmethod
    async def all(cls) -> List['Model']:
        """Get all objects"""
        return await cls.select().execute()

    async def save(self, force_insert: bool = False, force_update: bool = False):
        """Save the object to database"""
        database = self._meta.database
        if not database:
            raise ConfigurationError("No database configured for this model")

        table_name = self._meta.table_name
        placeholder = database.get_placeholder()

        # Update auto_now fields
        for name, field in self._fields.items():
            if isinstance(field, (DateTimeField, DateField)):
                if field.auto_now:
                    self._data[name] = datetime.now()

        # Get primary key field
        pk_field = None
        for name, field in self._fields.items():
            if field.primary_key:
                pk_field = name
                break

        pk_value = self._data.get(pk_field) if pk_field else None

        if force_insert or (not self._is_saved and not pk_value):
            # INSERT
            fields_to_insert = []
            values = []
            
            for name, field in self._fields.items():
                if field.auto_increment and not self._data.get(name):
                    continue
                if name in self._data and self._data[name] is not None:
                    fields_to_insert.append(field.get_column_name())
                    values.append(field.to_db(self._data[name]))

            if not fields_to_insert:
                return

            placeholders = ', '.join([placeholder] * len(values))
            query = f"INSERT INTO {table_name} ({', '.join(fields_to_insert)}) VALUES ({placeholders})"
            
            # Handle PostgreSQL RETURNING clause
            if isinstance(database, type(database)) and database.__class__.__name__ == 'PostgreSQLDatabase':
                query += f" RETURNING {pk_field}"
                result = await database.fetch_one(query, tuple(values))
                if result and pk_field:
                    self._data[pk_field] = result[pk_field]
            else:
                await database.execute(query, tuple(values))
                if pk_field and not self._data.get(pk_field):
                    self._data[pk_field] = await database.get_last_insert_id()

        else:
            # UPDATE
            if not pk_value:
                raise ValueError("Cannot update object without primary key")

            fields_to_update = []
            values = []
            
            for name, field in self._fields.items():
                if field.primary_key:
                    continue
                if name in self._data:
                    fields_to_update.append(f"{field.get_column_name()} = {placeholder}")
                    values.append(field.to_db(self._data[name]))

            if not fields_to_update:
                return

            values.append(pk_value)
            query = f"UPDATE {table_name} SET {', '.join(fields_to_update)} WHERE {pk_field} = {placeholder}"
            await database.execute(query, tuple(values))

        self._is_saved = True

    async def delete(self):
        """Delete the object from database"""
        database = self._meta.database
        if not database:
            raise ConfigurationError("No database configured for this model")

        # Get primary key
        pk_field = None
        for name, field in self._fields.items():
            if field.primary_key:
                pk_field = name
                break

        if not pk_field or not self._data.get(pk_field):
            raise ValueError("Cannot delete object without primary key")

        table_name = self._meta.table_name
        placeholder = database.get_placeholder()
        query = f"DELETE FROM {table_name} WHERE {pk_field} = {placeholder}"
        
        await database.execute(query, (self._data[pk_field],))
        self._is_saved = False

    async def refresh(self):
        """Refresh object from database"""
        pk_field = None
        for name, field in self._fields.items():
            if field.primary_key:
                pk_field = name
                break

        if not pk_field or not self._data.get(pk_field):
            raise ValueError("Cannot refresh object without primary key")

        refreshed = await self.__class__.get(**{pk_field: self._data[pk_field]})
        self._data = refreshed._data
        self._is_saved = refreshed._is_saved

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary"""
        result = {}
        for name, field in self._fields.items():
            value = self._data.get(name)
            if value is not None:
                result[name] = field.to_python(value)
        return result

    @classmethod
    def _from_db(cls, row: Dict[str, Any]) -> 'Model':
        """Create instance from database row"""
        instance = cls.__new__(cls)
        instance._data = {}
        instance._is_saved = True
        
        for name, field in cls._fields.items():
            column_name = field.get_column_name()
            if column_name in row:
                instance._data[name] = field.to_python(row[column_name])
        
        return instance

    @classmethod
    async def create_table(cls, safe: bool = True):
        """Create table for this model"""
        database = cls._meta.database
        if not database:
            raise ConfigurationError("No database configured for this model")

        table_name = cls._meta.table_name
        database_type = database.__class__.__name__.replace('Database', '').lower()

        columns = []
        for name, field in cls._fields.items():
            column_def = f"{field.get_column_name()} {field.get_sql_type(database_type)}"
            
            if field.primary_key:
                column_def += " PRIMARY KEY"
                if field.auto_increment:
                    if database_type == 'sqlite':
                        column_def += " AUTOINCREMENT"
                    elif database_type == 'mysql':
                        column_def += " AUTO_INCREMENT"
                    elif database_type == 'postgresql':
                        column_def = f"{field.get_column_name()} SERIAL PRIMARY KEY"
            
            if field.unique and not field.primary_key:
                column_def += " UNIQUE"
            
            if not field.null and not field.primary_key:
                column_def += " NOT NULL"
            
            if field.default is not None and not field.auto_increment and not callable(field.default):
                if isinstance(field.default, str):
                    column_def += f" DEFAULT '{field.default}'"
                else:
                    column_def += f" DEFAULT {field.default}"

            columns.append(column_def)

        # Add foreign key constraints
        for name, field in cls._fields.items():
            if isinstance(field, ForeignKeyField):
                related_table = field.related_model._meta.table_name
                fk_constraint = f"FOREIGN KEY ({field.get_column_name()}) REFERENCES {related_table}(id) ON DELETE {field.on_delete} ON UPDATE {field.on_update}"
                columns.append(fk_constraint)

        safe_clause = "IF NOT EXISTS " if safe else ""
        query = f"CREATE TABLE {safe_clause}{table_name} ({', '.join(columns)})"
        
        await database.execute(query)

    @classmethod
    async def drop_table(cls, safe: bool = True):
        """Drop table for this model"""
        database = cls._meta.database
        if not database:
            raise ConfigurationError("No database configured for this model")

        table_name = cls._meta.table_name
        safe_clause = "IF EXISTS " if safe else ""
        query = f"DROP TABLE {safe_clause}{table_name}"
        
        await database.execute(query)

    def __repr__(self):
        pk_field = None
        for name, field in self._fields.items():
            if field.primary_key:
                pk_field = name
                break
        
        pk_value = self._data.get(pk_field) if pk_field else None
        return f"<{self.__class__.__name__}: {pk_value}>"


"""Field types for models"""

import json
from datetime import datetime, date, time
from typing import Any, Optional, Type, Callable


class Field:
    """Base field class"""

    def __init__(
        self,
        null: bool = False,
        default: Any = None,
        unique: bool = False,
        index: bool = False,
        primary_key: bool = False,
        auto_increment: bool = False,
        column_name: Optional[str] = None,
        verbose_name: Optional[str] = None,
        help_text: Optional[str] = None,
    ):
        self.null = null
        self.default = default
        self.unique = unique
        self.index = index
        self.primary_key = primary_key
        self.auto_increment = auto_increment
        self.column_name = column_name
        self.verbose_name = verbose_name
        self.help_text = help_text
        self.name = None  # Set by metaclass
        self.model = None  # Set by metaclass

    def get_column_name(self) -> str:
        """Get database column name"""
        return self.column_name or self.name

    def to_python(self, value: Any) -> Any:
        """Convert database value to Python value"""
        if value is None:
            return None
        return value

    def to_db(self, value: Any) -> Any:
        """Convert Python value to database value"""
        if value is None:
            return None
        return value

    def get_sql_type(self, database_type: str) -> str:
        """Get SQL type for this field"""
        raise NotImplementedError

    def get_default(self) -> Any:
        """Get default value"""
        if callable(self.default):
            return self.default()
        return self.default


class CharField(Field):
    """Character field"""

    def __init__(self, max_length: int = 255, **kwargs):
        super().__init__(**kwargs)
        self.max_length = max_length

    def get_sql_type(self, database_type: str) -> str:
        return f"VARCHAR({self.max_length})"


class TextField(Field):
    """Text field"""

    def get_sql_type(self, database_type: str) -> str:
        return "TEXT"


class IntField(Field):
    """Integer field"""

    def to_python(self, value: Any) -> Optional[int]:
        if value is None:
            return None
        return int(value)

    def get_sql_type(self, database_type: str) -> str:
        return "INTEGER"


class BigIntField(Field):
    """Big integer field"""

    def to_python(self, value: Any) -> Optional[int]:
        if value is None:
            return None
        return int(value)

    def get_sql_type(self, database_type: str) -> str:
        return "BIGINT"


class FloatField(Field):
    """Float field"""

    def to_python(self, value: Any) -> Optional[float]:
        if value is None:
            return None
        return float(value)

    def get_sql_type(self, database_type: str) -> str:
        return "REAL"


class BoolField(Field):
    """Boolean field"""

    def to_python(self, value: Any) -> Optional[bool]:
        if value is None:
            return None
        if isinstance(value, bool):
            return value
        return value in (1, '1', 'true', 'True', 'TRUE', 't', 'T')

    def to_db(self, value: Any) -> Any:
        if value is None:
            return None
        return 1 if value else 0

    def get_sql_type(self, database_type: str) -> str:
        if database_type == "postgresql":
            return "BOOLEAN"
        return "INTEGER"


class DateField(Field):
    """Date field"""

    def __init__(self, auto_now: bool = False, auto_now_add: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.auto_now = auto_now
        self.auto_now_add = auto_now_add

    def to_python(self, value: Any) -> Optional[date]:
        if value is None:
            return None
        if isinstance(value, date):
            return value
        if isinstance(value, str):
            return datetime.fromisoformat(value).date()
        return value

    def to_db(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, date):
            return value.isoformat()
        return value

    def get_sql_type(self, database_type: str) -> str:
        return "DATE"


class DateTimeField(Field):
    """DateTime field"""

    def __init__(self, auto_now: bool = False, auto_now_add: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.auto_now = auto_now
        self.auto_now_add = auto_now_add
        if auto_now_add and not self.default:
            self.default = datetime.now

    def to_python(self, value: Any) -> Optional[datetime]:
        if value is None:
            return None
        if isinstance(value, datetime):
            return value
        if isinstance(value, str):
            return datetime.fromisoformat(value)
        return value

    def to_db(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, datetime):
            return value.isoformat()
        return value

    def get_sql_type(self, database_type: str) -> str:
        if database_type == "postgresql":
            return "TIMESTAMP"
        return "DATETIME"


class TimeField(Field):
    """Time field"""

    def to_python(self, value: Any) -> Optional[time]:
        if value is None:
            return None
        if isinstance(value, time):
            return value
        if isinstance(value, str):
            return datetime.fromisoformat(f"2000-01-01 {value}").time()
        return value

    def to_db(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, time):
            return value.isoformat()
        return value

    def get_sql_type(self, database_type: str) -> str:
        return "TIME"


class JSONField(Field):
    """JSON field"""

    def to_python(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, str):
            return json.loads(value)
        return value

    def to_db(self, value: Any) -> Any:
        if value is None:
            return None
        return json.dumps(value)

    def get_sql_type(self, database_type: str) -> str:
        if database_type == "postgresql":
            return "JSONB"
        return "TEXT"


class ForeignKeyField(Field):
    """Foreign key field"""

    def __init__(self, model: Type['Model'], on_delete: str = "CASCADE",
                 on_update: str = "CASCADE", **kwargs):
        super().__init__(**kwargs)
        self.related_model = model
        self.on_delete = on_delete
        self.on_update = on_update

    def get_sql_type(self, database_type: str) -> str:
        return "INTEGER"

    def get_column_name(self) -> str:
        """Get column name with _id suffix"""
        base_name = self.column_name or self.name
        if not base_name.endswith('_id'):
            base_name += '_id'
        return base_name